#include <QPainter>
#include <QDebug>

#include <linux/videodev2.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdint.h>
#include "v4l2_camera_view.h"
#include "yuyv_to_rgb.h"
#include "config.h"

V4l2CameraView::V4l2CameraView(QWidget *parent)
: QWidget(parent), m_numberOfBuffers(NUM_BUF)
{
	setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
	m_updateTimer = new QTimer(this);
	connect(m_updateTimer, SIGNAL(timeout()), this, SLOT(capture()));
}

V4l2CameraView::~V4l2CameraView()
{
	closeDevice();
}

void V4l2CameraView::setDeviceName(const QString &newDeviceName)
{
	if (m_deviceName != newDeviceName)
		m_deviceName = newDeviceName;
}

void V4l2CameraView::setFps(int newFps)
{
	if (m_fps != newFps) {
		m_fps = newFps;
		m_updateTimer->setInterval(1000/m_fps);
	}
}

int V4l2CameraView::start()
{
	int ret;
	m_imgConvertBuf = (uchar*)malloc(m_cameraWidth * m_cameraHeight * 4);
	
	if (m_imgConvertBuf == NULL) {
		qWarning() << __func__ << "malloc error";
		return -1;	
	}

	resize(m_cameraWidth, m_cameraHeight);
	for (int i = 0; i < m_numberOfBuffers; i++)
		m_images += QImage(size(), QImage::Format_RGB888);
	m_showing_image = 2;
	m_ready_image = -1;
	m_fd = -1;

	setDeviceName(DEV_VIDEO_PATH);
	ret = openDevice();
	if (ret) {
		qWarning() << __func__ << "opendevice error";
		return ret;
	}
	m_updateTimer->start();
	return 0;
}

void V4l2CameraView::stop()
{
	free(m_imgConvertBuf);
	m_updateTimer->stop();
	closeDevice();
}

int V4l2CameraView::setResolution(int width, int height)
{
	if (width <= 0 || height <= 0)
	       return -1;

	if (width != m_cameraWidth)
		m_cameraWidth = width;

	if (height != m_cameraHeight)
		m_cameraHeight = height;

	return 0;
}

void V4l2CameraView::paintEvent(QPaintEvent *)
{
	QPainter painter(this);
	if ((m_ready_image != -1) &&
	    (m_showing_image != m_ready_image)) {
		painter.drawImage(QPoint(0, 0),
				  m_images[m_ready_image]);
		m_showing_image = m_ready_image;
	}
}


void V4l2CameraView::capture()
{
	struct v4l2_buffer buf;
	int ret;

	memset(&buf, 0, sizeof(buf));
	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;
	
	if (m_fd == -1) {
		qWarning() << "m_fd is invalid" << __func__;
		return;
	}

	ret = ioctl(m_fd, VIDIOC_DQBUF, &buf);
	if (ret < 0) {
		qWarning() << "failed to de-queue buffer";
		return;
	}
	
	yuyv_to_rgb((uchar*)m_maps[buf.index], m_imgConvertBuf, m_cameraWidth * m_cameraHeight);
	m_images[buf.index] = QImage(m_imgConvertBuf, m_cameraWidth, m_cameraHeight,
			QImage::Format_RGB888);

	QPainter painter(&m_images[buf.index]);
	painter.setPen(Qt::white);
	painter.drawLine(0, m_cameraHeight/2, m_cameraWidth, m_cameraHeight/2);
	painter.drawLine(m_cameraWidth/2, 0, m_cameraWidth/2, m_cameraHeight);

	m_ready_image = buf.index;

	update();

	int unused_image = m_showing_image - 1;
	if (unused_image == -1)
		unused_image = m_numberOfBuffers - 1;

	memset(&buf, 0, sizeof(buf));
	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;
	buf.index = unused_image;

	if (m_fd == -1) {
		qWarning() << "m_fd is invalid" << __func__;
		return;
	}

	ret = ioctl(m_fd, VIDIOC_QBUF, &buf);
	if (ret < 0) {
		qWarning() << "failed to re-queue buffer " << buf.index;
		return;
	}
}

static int xioctl(int fd, int request, void *arg)
{
	for (; ; ) {
		int ret = ioctl(fd, request, arg);
		if (ret < 0) {
			if (errno == EINTR)
				continue;
			return -errno;
		}
		break;
	}
	return 0;
}

int V4l2CameraView::openDevice()
{
	struct v4l2_format fmt;
	struct v4l2_requestbuffers req;
	struct v4l2_buffer buf;
	enum v4l2_buf_type type;
	int ret;
	unsigned int i;

	closeDevice();

	m_fd = open(m_deviceName.toLocal8Bit().constData(), O_RDWR);
	if (m_fd == -1) {
		qWarning() << "failed to open " << m_deviceName;
		return -1;
	}

	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt.fmt.pix.width = m_cameraWidth;
	fmt.fmt.pix.height = m_cameraHeight;
	fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
	//fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24; /* EINVAL */
	//fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24; /* EINVAL */
	//fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;/* EINVAL */
	fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

	ret = xioctl(m_fd, VIDIOC_S_FMT, &fmt);
	if (ret < 0) {
		qWarning() << "failed to set format" << ret << m_deviceName.toLocal8Bit().constData() << m_fd << errno;
		goto err;
	}

	memset(&req, 0, sizeof(req));
	req.count = m_numberOfBuffers;
	req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory = V4L2_MEMORY_MMAP;

	ret = ioctl(m_fd, VIDIOC_REQBUFS, &req);
	if (ret < 0) {
		qWarning() << "failed to request buffers" << ret << strerror(errno);
		goto err;
	}

	for (i = 0; i < req.count; i++) {
		void *p;

		memset(&buf, 0, sizeof(buf));
		buf.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index  = i;
		ret = ioctl(m_fd, VIDIOC_QUERYBUF, &buf);
		if (ret < 0) {
			qWarning() << "failed to query buffer " << i;
			goto err;
		}

		p = mmap(NULL, buf.length, PROT_READ, MAP_SHARED,
			 m_fd, buf.m.offset);
		if (p == MAP_FAILED) {
			qWarning() << "failed to map the buffer " << i;
			goto err;
		}

		m_maps[i] = p;
		m_maps_len[i] = buf.length;
	}

	memset(&buf, 0, sizeof(buf));
	buf.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;
	buf.index  = 0;
	ret = ioctl(m_fd, VIDIOC_QBUF, &buf);
	if (ret < 0) {
		qWarning() << "failed to queue buffer " << i;
		goto err;
	}

	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	ret = ioctl(m_fd, VIDIOC_STREAMON, &type);
	if (ret < 0) {
		qWarning() << "failed to stream on";
		goto err;
	}

    return 0;
err:
	closeDevice();
	return -1;
}

void V4l2CameraView::closeDevice()
{
	int err;
	enum v4l2_buf_type type;
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (m_fd == -1)
		return;

	err = xioctl(m_fd, VIDIOC_STREAMOFF, &type);
	if (err != 0)
		qWarning() << "failed to VIDIOC_STREAMOFF" << "m_fd=" << m_fd << "err=" << err << "errno=" << errno;

	if (m_fd != -1) {
		for(int i=0; i < NUM_BUF; i++) {
			if (m_maps[i] != NULL)
				munmap(m_maps[i], m_maps_len[i]);
		}
		err = ::close(m_fd);
		if (err != 0)
			qWarning() << "failed to close" << m_fd << "errno=" << errno;

		m_fd = -1;
	}
}
